/* 
 * Copyright 2000 by Hans Reiser, licensing governed by reiserfs/README
 */
 
/*  
 * Written by Alexander Zarochentcev.
 * 
 * FS resize utility 
 *
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <asm/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/vfs.h>
#include <time.h>
#include <sys/ioctl.h>
#include <sys/mount.h>


#include "vfs.h"
#include "misc.h"
#include "reiserfs_fs.h"
#include "reiserfs_fs_sb.h"
#include "reiserfs.h"
#include "resize.h"

int opt_force = 0;
int opt_verbose = 1;			/* now "verbose" option is default */
int opt_nowrite = 0;
int opt_safe = 0;

/* Given a file descriptor and an offset, check whether the offset is
   a valid offset for the file - return 0 if it isn't valid or 1 if it
   is */
int valid_offset( int fd, loff_t offset )
{
  char ch;

  if (reiserfs_llseek (fd, offset, 0) < 0)
    return 0;

  if (read (fd, &ch, 1) < 1)
    return 0;

  return 1;
}


/* calculate the new fs size (in blocks) from old fs size and the string
   representation of new size */
static unsigned long calc_new_fs_size(unsigned long count, int bs, char *bytes_str) {
	long long int bytes;
	unsigned long blocks;
	int c;
	
	bytes = atoll(bytes_str);
	c = bytes_str[strlen(bytes_str) - 1];

	switch (c) {
	case 'G':
	    bytes *= 1024;
	case 'M':
	case 'm':
	    bytes *= 1024;
	case 'K':
	case 'k':
	    bytes *= 1024;
	}
	
	blocks = bytes / bs;

	if (bytes_str[0] == '+' || bytes_str[0] == '-')
		return (count + blocks);

	return blocks;
}

/* print some fs parameters */
static void sb_report(struct reiserfs_super_block * sb1,
       			      struct reiserfs_super_block * sb2){
	printf(
		"ReiserFS report:\n"
		"blocksize             %d\n"
		"block count           %d (%d)\n"
		"free blocks           %d (%d)\n"
		"bitmap block count    %d (%d)\n", 
		le16_to_cpu(sb1->s_blocksize),
		le32_to_cpu(sb1->s_block_count), le32_to_cpu(sb2->s_block_count),
		le32_to_cpu(sb1->s_free_blocks), le32_to_cpu(sb2->s_free_blocks),
		le16_to_cpu(sb1->s_bmap_nr), le16_to_cpu(sb2->s_bmap_nr));
};

/* read i-th bitmap block */
static struct buffer_head * get_bm_blk (int dev, int ind, int bs) {
	if (ind == 0) 
		return bread(dev, REISERFS_DISK_OFFSET_IN_BYTES / bs + 1 ,bs);
	return bread(dev, ind * bs * 8, bs);
}

/* conditional bwrite */
static int bwrite_cond (struct buffer_head * bh) {
	if(!opt_nowrite) { 
		mark_buffer_uptodate(bh,1);
		mark_buffer_dirty(bh,0);
		bwrite(bh);
	}
	return 0;
}


/* the first one of the mainest functions */
int expand_fs (struct super_block * s, unsigned long block_count_new) {
    struct buffer_head * bm_bh;
    int block_r, block_r_new;
    unsigned int bmap_nr_new;
    int i;
    
    /* count used bits in last bitmap block */
    block_r = SB_BLOCK_COUNT(s) - ((SB_BMAP_NR(s) - 1) * s->s_blocksize * 8);

    /* count bitmap blocks in new fs */
    bmap_nr_new = (block_count_new - 1) / (s->s_blocksize * 8) + 1;
    block_r_new = block_count_new - (bmap_nr_new - 1) * s->s_blocksize * 8;
    
    /* clear bits in last bitmap block (old layout) */
    bm_bh = get_bm_blk(s->s_dev, SB_BMAP_NR(s) - 1, s->s_blocksize);
    for (i = block_r; i < s->s_blocksize * 8; i++)
	clear_bit(i, bm_bh->b_data);
    bwrite_cond(bm_bh);
    brelse(bm_bh);
    
    /* add new bitmap blocks */
    for (i = SB_BMAP_NR(s); i < bmap_nr_new; i++) {
	bm_bh = get_bm_blk(s->s_dev, i, s->s_blocksize);
	memset(bm_bh->b_data, 0, bm_bh->b_size);
	set_bit(0, bm_bh->b_data);
	bwrite_cond(bm_bh);
	brelse(bm_bh);
    }
    
    /* set unused bits in last bitmap block (new layout) */
    bm_bh = get_bm_blk(s->s_dev, bmap_nr_new - 1, s->s_blocksize);
    for (i = block_r_new; i < s->s_blocksize * 8; i++)
	set_bit(i, bm_bh->b_data);
    bwrite_cond(bm_bh);
    brelse(bm_bh);
    
    /* update super block buffer*/
    PUT_SB_FREE_BLOCKS(s, SB_FREE_BLOCKS(s) + block_count_new - SB_BLOCK_COUNT(s) - (bmap_nr_new - SB_BMAP_NR(s)));
    PUT_SB_BLOCK_COUNT(s, block_count_new);
    PUT_SB_BMAP_NR(s, bmap_nr_new);
    
    /* commit changes */
    bwrite_cond(SB_BUFFER_WITH_SB(s));

    return 0;
}

int main(int argc, char *argv[]) {
	char * bytes_count_str = NULL;
	char * devname;
	struct stat statbuf;
	int c;

	int dev;
	struct super_block * s;
	struct reiserfs_super_block *sb_old;

	unsigned long block_count_new;
	
	printf ("\n%s\n", reiserfs_get_version_string());
	
	while ((c = getopt(argc, argv, "fvcqs:")) != EOF) {
		switch (c) {
		case 's' :
			  if (!optarg) 
				  die("%s: Missing argument to -s option", argv[0]);		
			  bytes_count_str = optarg;
			  break;
		case 'f':
		    opt_force = 1;
		    break;		 
		case 'v':
			opt_verbose++; 
			break;
		case 'n':
			/* no nowrite option at this moment */
			/* opt_nowrite = 1; */
			break;
		case 'c':
			opt_safe = 1;
			break;
		case 'q':
			opt_verbose = 0;
			break;
		default:
			print_usage_and_exit ();
		}
	}

	if (optind == argc )
		print_usage_and_exit();
	devname = argv[optind];

	/* open_device will die if it could not open device */
	dev = open (devname, O_RDWR);
	if (dev == -1)
		die ("%s: can not open '%s': %s", argv[0], devname, strerror (errno));

	if (fstat (dev, &statbuf) < 0)
		die ("%s: unable to stat %s", argv[0], devname);
  
	if (!S_ISBLK (statbuf.st_mode) && !opt_force )
		die ("%s: '%s (%o)' is not a block device", 
			 argv[0], devname, statbuf.st_mode);

	/* vfs superblock initialization */
	s = getmem (sizeof(struct super_block));
	s->s_dev = dev;
	uread_super_block(s);

	if (!new_format(s)) {
	    die ("resize_reiserfs: cannot resize reiserfs in old (groupped bitmap) format.\n");
	}

	if(bytes_count_str) {	/* new fs size is specified by user */
	    block_count_new = calc_new_fs_size(SB_BLOCK_COUNT(s), s->s_blocksize, bytes_count_str);
	} else {		/* use whole device */
	    loff_t dev_size;
	    dev_size = reiserfs_llseek(s->s_dev, (loff_t)0, SEEK_END);
	    if (dev_size == -1)
		die ("resize_reiserfs: cannot determine fs size.\n");
	    block_count_new = dev_size / s->s_blocksize;
	}

	if (is_mounted (devname)) {
		close(dev);
		if (!opt_force) 
		    die ("%s: '%s' contains a mounted file system,\n"
			 "\tspecify -f option to resize the fs online\n", 
			 argv[0], devname);
		return resize_fs_online(devname, block_count_new);
	}	
	
	if (SB_REISERFS_STATE(s) != REISERFS_VALID_FS) 
		die ("%s: the file system isn't in valid state\n", argv[0]);
		
	if(!valid_offset(dev, (loff_t) block_count_new * s->s_blocksize - 1))
		die ("%s: %s too small", argv[0], devname);

	sb_old = 0;		/* Needed to keep idiot compiler from issuing false warning */
	/* save SB for reporting */
	if(opt_verbose) {
	    sb_old = getmem(SB_SIZE);
	    memcpy(sb_old, SB_DISK_SUPER_BLOCK(s), SB_SIZE);
	}

	if (block_count_new == SB_BLOCK_COUNT(s)) 
	    die ("%s: Calculated fs size is the same as the previous one.", argv[0]);

	PUT_SB_REISERFS_STATE(s, REISERFS_ERROR_FS);
	bwrite_cond(SB_BUFFER_WITH_SB(s));	
	
	if (block_count_new > SB_BLOCK_COUNT(s)) 
	    expand_fs(s, block_count_new);
	else
	    shrink_fs(s, block_count_new);

	if(opt_verbose) {
		sb_report(SB_DISK_SUPER_BLOCK(s), sb_old);
		freemem(sb_old);
	}

	PUT_SB_REISERFS_STATE(s, REISERFS_VALID_FS);
	bwrite_cond(SB_BUFFER_WITH_SB(s));
	brelse(SB_BUFFER_WITH_SB(s));

	fsync_dev (0);
	
	if (opt_verbose) {
		printf("\nSyncing..");
		fflush(stdout);
	}
	fsync (dev);
	if (opt_verbose)
		printf("done\n");
	
	close(dev);
	
	return 0;
}
